'''
EN.640.635 Software Carpentry
Lab 2 - Solutions

In this lab assignment, we want to write two functions that can manipulate
images: one to blur the image and one to set the luminance (or brightness) of
the image.
'''
from PIL import Image
import numpy as np


def blur(fptr, mask=3):
    '''
    Apply a blur to an image. Saves both the original image and the
    newly-blurred image.

    **Parameters**

        fptr: *str*
            The name of an image file, with its extension
            (ex. spring.jpg, cat.png).
        mask: *int, optional*
            The size of our kernel mask.

    **Returns**

        None

    For extra info:

        *https://www.youtube.com/watch?v=C_zFhWdM4ic
    '''

    # We can open two images seperately, and only change img (thus, keeping
    # the original image unchanged for calculation purposes).
    # Note - we convert to RGB to circumvent some issues with different
    # file formats.
    original_img = Image.open(fptr).convert("RGB")
    img = Image.open(fptr).convert("RGB")
    width, height = img.size

    # To make sure we get the correct mask size, take our input dimension,
    # divide by 2 and cast to an integer to get the proper dimension.
    mask = int(mask / 2)

    # Parse through image
    print("Blurring Image...")
    for x in range(width):
        for y in range(height):
            # Get all neighboring pixels
            total_pxl_color_in_mask = np.array([0, 0, 0])
            count = 0
            for xi in range(-mask + x, mask + 1 + x):
                for yi in range(-mask + y, mask + 1 + y):
                    if xi < 0 or yi < 0 or xi >= width or yi >= height:
                        # If we get here, we are out of bounds
                        continue
                    # If we get here, we are in bounds
                    total_pxl_color_in_mask += np.array(img.getpixel((xi, yi)))
                    count += 1
            # Find the pixel blur, and assign it.
            blur = total_pxl_color_in_mask / count
            blur = tuple(map(int, blur))
            img.putpixel((x, y), blur)

    # Save both images so we can verify if we changed the correct one.
    base_name = '.'.join(fptr.split(".")[0:-1])
    fptr_2 = base_name + "_blurred.png"
    img.save(fptr_2)
    fptr_2 = base_name + "_original.png"
    original_img.save(fptr_2)
    print("Image has been blurred.\n")


def set_luminance_1(fptr, l_val, max_iter=1000000, eps=1.0):
    '''
    Luminance is a method of determining how "bright" the image is. It can
    be easily calculated per pixel with the following formula:

        l_pxl = 0.299 * R + 0.587 * G + 0.114 * B

    Now, for an entire image we can calculate either the total luminance, or
    the average luminance over the whole image:

        l_avg = sum(l_pxl) / N_pxl

    We want to set the luminance of an image to a user specified value,
    allowing us to essentially set how bright our image is. We can use the
    get_pxl_luminance() and get_luminance() functions below to help us.

    **Parameters**

        fptr: *str*
            The name of an image file, with its extension
            (ex. spring.jpg, cat.png).
        l_val: *int*
            The desired luminance to set the image to.
        max_iter: *int, optional*
            The maximum number of iterations for adjusting luminance.
        eps: *float, optional*
            Error tolerance for determining convergence.

    **Returns**

        None
    '''
    print("Setting luminance...")
    # Need to check that the specified luminance value is between 0 and 255.
    l_val = float(l_val)
    assert 0.0 <= l_val <= 255.0, "Luminance must be between 0 and 255!"
    # Read image
    img = Image.open(fptr).convert("RGB")
    width, height = img.size
    # Initialize the original difference
    diff = l_val - get_luminance(img)
    for i in range(max_iter):
        # Adjust the pixels of the image by the difference between the
        # two values. After several iterations, the calculated
        # luminance value will converge at the desired value.
        for x in range(width):
            for y in range(height):
                current_colors = np.array(img.getpixel((x, y)))
                new_colors = tuple(current_colors + int(diff))
                img.putpixel((x, y), new_colors)
        # Calculate new luminance and difference
        current_l = get_luminance(img)
        diff = l_val - current_l
        # Check if it is close enough to the specified luminance value
        if abs(diff) < eps:
            break

    # Save the new image
    base_name = '.'.join(fptr.split(".")[0:-1])
    fptr_2 = base_name + "_lumin.png"
    img.save(fptr_2)
    print("Luminance set to %.1f" % current_l)


def set_luminance_2(fptr, l_val):
    '''
    Luminance is a method of determining how "bright" the image is. It can
    be easily calculated per pixel with the following formula:

        l_pxl = 0.299 * R + 0.587 * G + 0.114 * B

    Now, for an entire image we can calculate either the total luminance, or
    the average luminance over the whole image:

        l_avg = sum(l_pxl) / N_pxl

    We want to set the luminance of an image to a user specified value,
    allowing us to essentially set how bright our image is. We can use the
    get_pxl_luminance() and get_luminance() functions below to help us.

    **Parameters**

        fptr: *str*
            The name of an image file, with its extension
            (ex. spring.jpg, cat.png).
        l_val: *int*
            The desired luminance to set the image to.
        max_iter: *int, optional*
            The maximum number of iterations for adjusting luminance.
        eps: *float, optional*
            Error tolerance for determining convergence.

    **Returns**

        None
    '''
    img = Image.open(fptr).convert("RGB")
    width, height = img.size
    base_name = '.'.join(fptr.split(".")[0:-1])
    # The simplest approach to solving this problem is to find a total scalar
    # for the image that would increment the luminance directly to the target.
    # This approach has some failings though, as the max luminance of a pixel
    # is 255, meaning that if we have a pixel that is (255, 1, 0), and we
    # scale it by 200, the end result is (255, 200, 0), a vastly different
    # color!
    current_luminance = get_luminance(img)
    scalar = l_val / current_luminance
    for x in range(width):
        for y in range(height):
            pxl = scalar * np.array(img.getpixel((x, y)))
            img.putpixel(
                (x, y),
                tuple(map(lambda v: min(int(v), 255), pxl))
            )
    fptr_2 = base_name + "_lumin.png"
    img.save(fptr_2)


def get_pxl_luminance(pxl):
    '''
    Given a pixel, this function will calculate its luminance.

    **Parameters**

        pxl: *tuple, int*
            A tuple of integers holding RGB values.

    **Returns**

        l_val: *float*
            The pixel luminance.
    '''
    return 0.299 * pxl[0] + 0.587 * pxl[1] + 0.114 * pxl[2]


def get_luminance(img):
    '''
    Returns the average luminance of an image.

    **Parameters**

        img: *PIL.Image*
            A PIL image object.

    **Returns**

        l_val: *float*
            The image luminance.
    '''
    width, height = img.size
    all_luminance = [get_pxl_luminance(img.getpixel((x, y)))
                     for x in range(width)
                     for y in range(height)]
    return sum(all_luminance) / float(len(all_luminance))


if __name__ == "__main__":
    blur("gray.jpg")
    set_luminance_1("gray.jpg", 100)
    set_luminance_2("gray.jpg", 100)
